package CatCore;

import java.io.*;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

public class ReadPrologDatabase {
    
    // schema is a map of (tableName,array-list-of-column-names)
    // tables is a map of (tableName, list-of-tuples) where a tuple is a map (columnName,value) pairs

    public Map<String, List<Map<String, String>>> tables;
    public Map<String, List<String>> schema;
    public String fileName;
    
    // ReadPrologDatabase will exit if any error is invoked.  Otherwise, the
    // database itself can be accessed through tables and schema variables above.

    public ReadPrologDatabase(File infile, PrintStream eps) {
        boolean hasErrors = false;
        tables = new HashMap<>();
        schema = new HashMap<>();
        
        try {
            fileName = infile.getName();
            // Step 1: 1st pass for tables. Ignore all lines except those that start with '%table'
            LineNumberReader br = new LineNumberReader(new InputStreamReader(
                    new FileInputStream(infile)));
            String line;
            while ((line = br.readLine()) != null) {

                /* START NEW CODE */
                line = line.trim();
                if (line.startsWith("%table(")) { // parse table facts
                    StringTokenizer st = new StringTokenizer(line, " \t(),][.\"");
                    try {
                        st.nextToken(); // ignore "table"
                        List<String> columnNames = new ArrayList<>();
                        while (st.hasMoreTokens()) {
                            String name = st.nextToken().trim();
                            columnNames.add(name);
                        }
                        schema.put(columnNames.remove(0), columnNames);
                    } catch (Exception e) {
                        eps.format("unable to parse table on line %d in %s\n", 
                                br.getLineNumber(), infile.getName());
                        hasErrors=true;
                    }
                }
            }
            
            // Step 2: now create the tables map
            for (String tableName : schema.keySet()) {
                tables.put(tableName + "S",
                        new ArrayList<Map<String, String>>());
            }
            
            // Step 3: 2nd pass for table tuples.  Ignore lines that start with 
            // 'table(' or 'tuple(' or 'dbase('
            br = new LineNumberReader(new InputStreamReader(new FileInputStream(
                    infile)));
            while ((line = br.readLine()) != null) {
                line = line.trim();
                // ignore schema and tuple facts and assorted computations
                if (line.startsWith("%table(") || line.startsWith("tuple(") 
                    || line.startsWith("%dbase(") || line.startsWith(":- ")) {
                    continue;
                }
                
                // Step 3.1: we found a tuple line -- see if it passes minimal
                //           structure testing
                if (line.matches("\\s*\\w+\\s*(.*)\\.")) {
                    
                    // Step 3.1.1 : get the table name (in tableName)
                    //              the string inbetween '('...')' in strippedLine
                    //           
                    String tableName = line.split("\\(")[0].trim();
                    String strippedLine = line.substring(line.indexOf("(") + 1, line.lastIndexOf(")"));
                    if (strippedLine.equals(""))
                        continue;
                    
                    // Step 3.1.2: if we don't recognize the name of the table, ignore the line
                    if (!schema.containsKey(tableName)) {
                        continue;
                    }
                    
                    // Step 3.1.3: now produce a new string that allows us to split all of its
                    //             pieces by "@@@".
                    
                    String newline = "";
                    boolean inString = false;
                    for (int i=0; i<strippedLine.length(); i++) {
                        char current = strippedLine.charAt(i);
                        if (current == '\'') {
                            inString=!inString;
                        }
                        if (!inString && current == ',')
                            newline = newline + "@@@";
                        else
                            newline = newline + current;
                    }
                    
                    String[] columnValues = newline.split("@@@");
                    
                    // Step 3.1.4: get the current set of tuples for the given table in variable tuples
                    
                    List<Map<String, String>> tuples = tables.get(tableName + "S");
                    Map<String, String> newTuple = new HashMap<>();
                    List<String> tableDef = schema.get(tableName);
                    for (int i = 0; i < tableDef.size(); i++) {
                        String value = trimmed(columnValues[i]);
                        newTuple.put(tableDef.get(i), value);
                    }
                    tuples.add(newTuple);
                }
            }
        } catch (IOException e) {
            e.printStackTrace();
            hasErrors=true;
        }
        if (hasErrors)
            throw new ExecutionError("C20: errors in parsing prolog table");
    }
    
    private String trimmed(String s) {
        s = s.trim();
        if (s.startsWith("'")) {
            s = s.substring(1,s.length()-1);
        }
        return s;
    }
}
