package CVParser;

import CVParser.importer.ClassEdge;
import CVParser.importer.ClassInterface;
import CVParser.importer.ClassNode;
import CVParser.importer.ClassViolet;
import CVParser.importer.Importer;
import CVParser.importer.Node;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;

public class DBWriter {

    public static String DB_FILE_APPENSION = ".pl";
    String outputFile;
    ArrayList<ClassNode> classNodes;
    ArrayList<ClassEdge> classEdges;
    ArrayList<ClassInterface> classInterfaces;
    ClassViolet vData;
    int numNodes;
    boolean printedVioletInterfaceExtends;
    boolean printedVioletClassImplements;
    int counter = 0;

    public DBWriter(String outputFile, ArrayList<ClassNode> classNodes,
            ArrayList<ClassEdge> classEdges, ArrayList<ClassInterface> classInterfaces,
            ClassViolet vData, int numNodes) {
        this.outputFile = outputFile;
        this.classNodes = classNodes;
        this.classEdges = classEdges;
        this.classInterfaces = classInterfaces;
        this.vData = vData;
        this.numNodes = numNodes;
        this.printedVioletInterfaceExtends = false;
        this.printedVioletClassImplements = false;
    }

    public void generateVioletDB() {
        PrintStream out;
        try {
            out = new PrintStream(outputFile);

            out.format("%s", "\n%table(vClass,[id,\"name\",x,y]).\n");
            if (classNodes.size() > 0) {
                for (ClassNode node : classNodes) {
                    writeClassTuples(node, out);
                }
            } else {
                out.format("%s", ":- dynamic vClass/4.\n");
            }

            //Output the interfaces
            out.format("%s", "\n%table(vInterface,[id,\"name\",x,y]).\n");
            if (classInterfaces.size() > 0) {
                for (ClassInterface node : classInterfaces) {
                    writeInterfaceTuples(node, out);
                }
            } else {
                out.format("%s", ":- dynamic vInterface/4.\n");
            }

            // output fields
            out.format("%s", "\n%table(vField,[id,\"name\",\"type\",cid]).\n");
            fieldCtr = 0;
            for (ClassNode node : classNodes) {
                writeFieldTuples(node, out);
            }
            if (fieldCtr==0) {
                out.format("%s", ":- dynamic vField/4.\n");
            }

            out.format("%s", "\n%table(vMethod,[id,\"name\",cid]).\n");
            methCtr = 0;
            for (Node n : classNodes) {
                writeMethodTuples(n, out);
            }
            for (Node nd : classInterfaces) {
                writeMethodTuples(nd, out);
            }
            if (methCtr == 0) {
                out.format("%s", ":- dynamic vMethod/3.\n");
            }

            //Output the associations
            //counter++, startId, startType, startCard, startArrow, endId, endType, endCard, endArrow, endType, lineStyle
            out.format("%s", "\n%table(vAssoc,[id,\"name\",cid1,type1,\"card1\",\"arrow1\",cid2,type2,\"card2\",\"arrow2\",\"lineStyle\").\n");
            int yyy = classEdges.size();
            if (classEdges.size() > 0) {
                for (ClassEdge edge : classEdges) {
                    writeAssociationTuples(edge, out);
                }
            } else {
                out.format("%s", ":- dynamic vAssoc/11.\n");
            }
            
            out.close();
        } catch (IOException e) {
            System.err.println(e.getMessage());
            e.printStackTrace();
        }
    }

    private static void writeInterfaceTuples(ClassInterface node, PrintStream out) throws IOException {
        String id = node.getId();
        String name = node.getName();
        name = name.replace("«interface»", "").replace("interface", "").trim();  // Added by *DSB*
        out.format("vInterface(%s,'%s',%d,%d).\n", id, name, node.getXPos(), node.getYPos());
    }

    private static void writeClassTuples(ClassNode node, PrintStream out) throws IOException {
        String id = node.getId();
        String name = formatAttrsAndMeths(node.getName().replace("«interface»", "").trim());
        out.format("vClass(%s,'%s',%d,%d).\n", id, name, node.getXPos(), node.getYPos());
    }

    private static int fieldCtr;

    private static void writeFieldTuples(ClassNode node, PrintStream out) throws IOException {
        String type = null; // pacify whiney compiler
        String id = node.getId();
        String name = formatAttrsAndMeths(node.getName().replace("«interface»", "").trim());
        String attrs = formatAttrsAndMeths(node.getAttributes());
        if (attrs.equals("")) {
            return;
        }

        String[] attrList = attrs.split(";");
        for (String field : attrList) {
            field = field.trim();
            if (field.contains("'")) {
                type = "String";
            } else if (field.contains("?")) {
                type = "boolean";
            } else {
                type = "int";
            }
            name = field.replace("'", "").replace("?","");
            out.format("vField(f%d,'%s','%s',%s).\n", fieldCtr++, name, type, id);
        }
    }

    private static int methCtr;

    private static void writeMethodTuples(Node node, PrintStream out) throws IOException {
        String type = null; // pacify whiney compiler
        String id = node.getId();
        String name = formatAttrsAndMeths(node.getName().replace("«interface»", "").trim());
        String methods = formatAttrsAndMeths(node.getMethods());
        if (methods.equals("")) {
            return;
        }

        String[] methList = methods.split(";");
        for (String meth : methList) {
            meth = meth.trim();
            if (meth.contains("'")) {
                type = "String";
            } else if (meth.contains("?")) {
                type = "boolean";
            } else {
                type = "int";
            }
            name = meth.replace("'", "").replace("?","");
            out.format("vMethod(m%d,'%s','%s',%s).\n", methCtr++, name, type, id);
        }
    }

    private void writeAssociationTuples(ClassEdge edge, PrintStream out) throws IOException {
        String startId = edge.getStartsAtClass();
        String endId = edge.getEndsAtClass();
        String startArrow = formatTextVariable(edge.getStartArrowHead());
        String endArrow = formatTextVariable(edge.getEndArrowHead());
        String startCard = edge.getStartLabel();
        String endCard = edge.getEndLabel();
        String lineStyle = formatTextVariable(edge.getLineStyle());
        String startType = edge.getStartsType();
        String endType = edge.getEndsType();
        String middleLabel = formatTextVariable(edge.getMiddleLabel());

        startCard = startCard.replace(",", ";");
        startCard = formatTextVariable(Importer.allFirstCharsAreLowerCase(startCard, ";"));

        //startLabel = formatTextVariable(Importer.allFirstCharsAreLowerCase(startCard.replace(",", ";"), ";"));
        endCard = formatTextVariable(Importer.allFirstCharsAreLowerCase(endCard.replace(",", ";"), ";"));

        //Always need to write an association, regardless of edge type, for Violet purposes
        out.format("vAssoc(a%d,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s).\n",
                counter++, middleLabel, startId, startType, startCard, startArrow, endId, endType, endCard, endArrow, endType, lineStyle);
    }

    private static String formatId(String input) {
        return input != null ? input.replaceAll("[^0-9]", "") : "''";
    }

    private static String formatAttrsAndMeths(String input) {
        if (input == null || input.equals("[]")) {
            return "''";
        }
        String output = input.replace("[", "'");
        output = output.replace("]", "'");
        output = output.replace(", ", ";");

        return output;
    }

    private static String formatTextVariable(String input) {
        if (input.startsWith("'") && input.endsWith("'"))
            return input;
        if (input == null) {
            return "''";
        }
        return "'" + input + "'";
    }
}
