package CatLite.classVioletParser.writer;

import CatLite.classVioletParser.importer.ClassInterface;
import CatLite.classVioletParser.importer.ClassViolet;
import CatLite.classVioletParser.importer.ClassNode;
import CatLite.classVioletParser.importer.ClassEdge;
import CatLite.classVioletParser.importer.Importer;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;

public class DBWriter {

    public static String DB_FILE_APPENSION = ".pl";
    String outputFile;
    ArrayList<ClassNode> classNodes;
    ArrayList<ClassEdge> classEdges;
    ArrayList<ClassInterface> classInterfaces;
    ClassViolet violetData;
    int numNodes;
    boolean printedVioletInterfaceExtends;
    boolean printedVioletClassImplements;
    int counter = 0;

    public DBWriter(String outputFile, ArrayList<ClassNode> classNodes,
            ArrayList<ClassEdge> classEdges, ArrayList<ClassInterface> classInterfaces,
            ClassViolet violetData, int numNodes) {
        this.outputFile = outputFile;
        this.classNodes = classNodes;
        this.classEdges = classEdges;
        this.classInterfaces = classInterfaces;
        this.violetData = violetData;
        this.numNodes = numNodes;
        this.printedVioletInterfaceExtends = false;
        this.printedVioletClassImplements = false;
    }

    public void generateVioletDB() {
        try {
            FileWriter fstream = new FileWriter(outputFile);
            BufferedWriter out = new BufferedWriter(fstream);

            //Output special pl settings
            out.write(":- style_check(-discontiguous).\n");

            //Output header data
            //Ouput the classes
            out.write("\n%table(violetClass,[id,\"name\",\"fields\",\"methods\",x,y]).\n");
            if (classNodes.size() > 0) {
                for (ClassNode node : classNodes) {
                    writeClassTuples(node, out);
                }
            } else {
                out.write(":- dynamic violetClass/6.\n");
            }

            //Ouput the interfaces
            out.write("\n%table(violetInterface,[id,\"name\",\"methods\",x,y]).\n");
            if (classInterfaces.size() > 0) {
                for (ClassInterface node : classInterfaces) {
                    writeInterfaceTuples(node, out);
                }
            } else {
                out.write(":- dynamic violetInterface/5.\n");
            }

            //Output the associations
            out.write("\n%table(violetAssociation,[id,\"role1\",\"arrow1\",type1,\"role2\",\"arrow2\",type2,\"lineStyle\",cid1,cid2]).\n");
            int yyy = classEdges.size();
            if (classEdges.size() > 0) {
                for (ClassEdge edge : classEdges) {
                    writeAssociationTuples(edge, out);
                }
            } else {
                out.write(":- dynamic violetAssociation/10.\n");
            }

            out.write("\n%table(violetMiddleLabels,[id,cid1,cid2,\"label\"]).\n");
            boolean someTuples = false;
            boolean result;

            for (ClassEdge edge : classEdges) {
                result = formatAndWriteMiddleFacts(edge, out);
                someTuples = result || someTuples;
            }
            if (!someTuples) {
                out.write(":- dynamic violetMiddleLabels/4.\n");
            }
            out.close();
        } catch (IOException e) {
            System.err.println(e.getMessage());
            e.printStackTrace();
        }
    }

    private static void writeInterfaceTuples(ClassInterface node, BufferedWriter out) throws IOException {
        String id = node.getId();
        String name = node.getName();
        // Violet automatically places <<inteface>> as part of the name;
        // this and any embedded blanks should be stripped.
        name = name.replace("«interface»", "").replace("interface","").trim();  // Added by *DSB*
        String methods = formatAttrsAndMeths(node.getMethods().toString());

        out.write("violetInterface(" + id + ",'" + name + "'," + methods + ","
                + node.getXPos() + "," + node.getYPos() + ").\n");
    }

    private static void writeClassTuples(ClassNode node, BufferedWriter out) throws IOException {
        String id = node.getId();
        String name = formatAttrsAndMeths(node.getName().replace("«interface»", "").trim());
        String attrs = formatAttrsAndMeths(node.getAttributes().toString());
        String methods = formatAttrsAndMeths(node.getMethods().toString());

        attrs = attrs.replace("'", "");
        attrs = Importer.allFirstCharsAreLowerCase(attrs, ";");

        out.write("violetClass(" + id + ",'" + name + "','" + attrs + "',"
                + methods + "," + node.getXPos() + "," + node.getYPos() + ").\n");
    }

    private void writeAssociationTuples(ClassEdge edge, BufferedWriter out) throws IOException {
        String startId = edge.getStartsAtClass();
        String endId = edge.getEndsAtClass();
        String startArrow = formatTextVariable(edge.getStartArrowHead());
        String endArrow = formatTextVariable(edge.getEndArrowHead());
        String startLabel = edge.getStartLabel();
        String endLabel = edge.getEndLabel();
        String lineStyle = formatTextVariable(edge.getLineStyle());
        String startsType = edge.getStartsType();
        String endsType = edge.getEndsType();

        startLabel = startLabel.replace(",", ";");
        startLabel = formatTextVariable(Importer.allFirstCharsAreLowerCase(startLabel, ";"));

        //startLabel = formatTextVariable(Importer.allFirstCharsAreLowerCase(startLabel.replace(",", ";"), ";"));
        endLabel = formatTextVariable(Importer.allFirstCharsAreLowerCase(endLabel.replace(",", ";"), ";"));

        //Always need to write an association, regardless of edge type, for Violet purposes
        out.write("violetAssociation(id" + counter++ + "," + startLabel + "," + startArrow
                + "," + startsType + "," + endLabel + "," + endArrow + "," + endsType + "," + lineStyle
                + "," + startId + "," + endId + ").\n");
    }


private boolean formatAndWriteMiddleFacts(ClassEdge edge, BufferedWriter out) throws IOException {
        String fromId = edge.getStartsAtClass();
        String toId = edge.getEndsAtClass();
        String middleLabel = formatTextVariable(edge.getMiddleLabel());

        if (middleLabel.equals("''")) {
            return false;
        }
        out.write("violetMiddleLabels(id" + counter++ + "," + fromId + "," + toId + "," + middleLabel + ").\n");
        return true;
    }

    private static String formatId(String input) {
        return input != null ? input.replaceAll("[^0-9]", "") : "''";
    }

    private static String formatAttrsAndMeths(String input) {
        if (input == null || input.equals("[]")) {
            return "''";
        }
        String output = input.replace("[", "'");
        output = output.replace("]", "'");
        output = output.replace(", ", ";");

        return output;
    }

    private static String formatTextVariable(String input) {
        if (input == null) {
            return "''";
        }
        return "'" + input + "'";
    }
}
